import subprocess
import re
import sys
import shutil
import os
from typing import List

import click


# Create a Click group
@click.group()
def cdbt():
    """cdbt is a CLI tool to manage dbt builds and handle states."""
    pass


@cdbt.command()
@click.option('--full-refresh', is_flag=True, help='Run a full refresh on all models.')
@click.option('--select', type=str, help='DBT style select string')
@click.option('--fail-fast', is_flag=True, help='Fail fast on errors.')
def build(full_refresh, select, fail_fast):
    args = []
    if select:
        args.append('--select')
        args.append(select)
    if fail_fast:
        args.append('--fail-fast')
    print(f'Starting a dbt build with args: {args}')
    try:
        run_result = execute_dbt_command('seed', args)
    except subprocess.CalledProcessError as e:
        print(e)
        sys.exit(1)

    if not run_result:
        raise DbtError('DBT build failed with errors.')

    run_args = args
    if full_refresh:
        run_args.append('--full-refresh')
    try:
        run_result = execute_dbt_command('seed', run_args)
    except subprocess.CalledProcessError as e:
        print(e)
        sys.exit(1)

    if not run_result:
        raise DbtError('DBT build failed with errors.')

    if run_result:

        try:
            run_result = execute_dbt_command('test', args)
        except subprocess.CalledProcessError as e:
            print(e)
            sys.exit(1)


@cdbt.command()
@click.option('--full-refresh', is_flag=True, help='Run a full refresh on all models.')
@click.option('--select', type=str, help='DBT style select string')
@click.option('--fail-fast', is_flag=True, help='Fail fast on errors.')
def trun(full_refresh, select, fail_fast):
    args = []
    if select:
        args.append('--select')
        args.append(select)
    if fail_fast:
        args.append('--fail-fast')
    print(f'Starting a dbt build with args: {args}')
    run_args = args
    if full_refresh:
        run_args.append('--full-refresh')
    try:
        run_result = execute_dbt_command('seed', run_args)
    except subprocess.CalledProcessError as e:
        print(e)
        sys.exit(1)

    if not run_result:
        raise DbtError('DBT build failed with errors.')

    if run_result:
        try:
            run_result = execute_dbt_command('test', args)
        except subprocess.CalledProcessError as e:
            print(e)
            sys.exit(1)


@cdbt.command()
@click.option('--full-refresh', is_flag=True, help='Run a full refresh on all models.')
@click.option('--select', type=str, help='DBT style select string')
@click.option('--fail-fast', is_flag=True, help='Fail fast on errors.')
def run(full_refresh, select, fail_fast):
    args = []
    if select:
        args.append('--select')
        args.append(select)
    if fail_fast:
        args.append('--fail-fast')

    run_args = args
    if full_refresh:
        run_args.append('--full-refresh')
    try:
        run_result = execute_dbt_command('seed', run_args)
    except subprocess.CalledProcessError as e:
        print(e)
        sys.exit(1)


@cdbt.command()
@click.option('--select', type=str, help='DBT style select string')
@click.option('--fail-fast', is_flag=True, help='Fail fast on errors.')
def test(full_refresh, select, fail_fast):
    args = []
    if select:
        args.append('--select')
        args.append(select)
    if fail_fast:
        args.append('--fail-fast')

    run_args = args
    if full_refresh:
        run_args.append('--full-refresh')
    try:
        run_result = execute_dbt_command('seed', run_args)
    except subprocess.CalledProcessError as e:
        print(e)
        sys.exit(1)


@cdbt.command()
def compile(full_refresh, select, fail_fast):
    try:
        subprocess.run(['dbt', 'compile'], check=True)
    except subprocess.CalledProcessError as e:
        print(e)
        sys.exit(1)


@cdbt.command()
@click.option('--full-refresh', is_flag=True, help='Run a full refresh on all models.')
def sbuild(full_refresh):
    print('Starting a state build based on local manifest.json')
    artifact_dir = '_artifacts'
    target_dir = 'target'
    # Path to the artifacts file that will be generated by the dbt compile command representing the current state.
    manifest_path = os.path.join('./', target_dir, 'manifest.json')
    # Path to the artifact file that represents the prior build state.
    manifest_artifact_path = os.path.join('./', artifact_dir, 'manifest.json')

    execute_state_based_build(artifact_dir, manifest_artifact_path, manifest_path, full_refresh,
                              roll_back_manifest_flag=True)


@cdbt.command()
@click.option('--full-refresh', is_flag=True, help='Run a full refresh on all models.')
def pbuild(full_refresh):
    print('Starting a state build based on production manifest.json')
    artifact_dir = 'logs'
    target_dir = 'target'
    # Pull artifacts from Snowflake. These are the latest production artifacts.
    try:
        subprocess.run(['dbt', 'run-operation', 'get_last_artifacts'], check=True)
    except subprocess.CalledProcessError as e:
        print(e)
        sys.exit(1)

    manifest_path = os.path.join('.', target_dir, 'manifest.json')
    manifest_artifact_path = os.path.join('.', artifact_dir, 'manifest.json')

    execute_state_based_build(artifact_dir, manifest_artifact_path, manifest_path, full_refresh,
                              roll_back_manifest_flag=False)


def execute_state_based_build(artifact_dir: str, manifest_artifact_path: str, manifest_path: str, full_refresh: bool,
                              roll_back_manifest_flag: bool):
    if roll_back_manifest_flag:
        print(f'Making a backup of the current manifest.json at {manifest_path} to {manifest_artifact_path}')
        # Move the manifest from ./target to ./_artifacts. This becomes the prior state. Only used for local state
        # build. Not used for pdbuild (production build).
        shutil.move(manifest_path, manifest_artifact_path)
    # Execute dbt compile
    try:
        subprocess.run(['dbt', 'compile'], check=True)
    except subprocess.CalledProcessError:
        sys.exit(1)
    # Construct state commands
    state_flags = [
        '--select', 'state:modified+',
        '--state', os.path.join('.', artifact_dir) + '/'
    ]

    # Execute dbt seed
    try:
        run_result = execute_dbt_command('seed', state_flags)
    except subprocess.CalledProcessError as e:
        if roll_back_manifest_flag:
            roll_back_manifest(e, manifest_artifact_path, manifest_path)
            sys.exit(1)
        else:
            print(e)
            sys.exit(1)

    if not run_result:
        e = 'DBT build failed with errors.'
        roll_back_manifest(e, manifest_artifact_path, manifest_path)
        raise DbtError('DBT build failed with errors.')

    # Execute dbt run
    if full_refresh:
        run_flags = state_flags + ['--full-refresh']
    else:
        run_flags = state_flags
    try:
        run_result = execute_dbt_command('run', run_flags)
    except subprocess.CalledProcessError as e:
        if roll_back_manifest_flag:
            roll_back_manifest(e, manifest_artifact_path, manifest_path)
        else:
            print(e)
            sys.exit(1)

    if not run_result:
        e = 'DBT build failed with errors.'
        roll_back_manifest(e, manifest_artifact_path, manifest_path)
        raise DbtError('DBT build failed with errors.')

    # Execute dbt test
    try:
        run_result = execute_dbt_command('test', state_flags)
    except subprocess.CalledProcessError as e:
        if roll_back_manifest_flag:
            roll_back_manifest(e, manifest_artifact_path, manifest_path)
        else:
            print(e)
            sys.exit(1)

    if not run_result:
        e = 'DBT build failed with errors.'
        roll_back_manifest(e, manifest_artifact_path, manifest_path)
        raise DbtError('DBT build failed with errors.')


def roll_back_manifest(e, manifest_artifact_path, manifest_path):
    print(f"DBT build failed. Rolling back manifest state with error\n {e}")
    # Move the manifest.json from _artifacts back to target dir. If the build failed, we want to rebuild against this
    # state, not the one generated by the compile command.
    shutil.move(manifest_artifact_path, manifest_path)
    sys.exit(1)


def execute_dbt_command(command: str, args: List[str]):
    dbt_command = ['dbt', command] + args
    print(f'Running command: {dbt_command}')

    process = subprocess.Popen(
        dbt_command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True  # Ensure outputs are in text mode rather than bytes
    )

    # Real-time output streaming
    while True:
        output = process.stdout.readline()
        if output == '' and process.poll() is not None:
            break
        if output:
            print(output.strip())  # Print each line of the output

    # Capture and print any remaining output after the loop
    stdout, stderr = process.communicate()
    if stdout:
        print(stdout.strip())

    # Check exit code
    if process.returncode != 0:
        print(f"Command resulted in an error: {stderr}")
        raise subprocess.CalledProcessError(returncode=process.returncode, cmd=dbt_command, output=stderr)

    # Check for errors using a regex method if necessary
    if contains_errors(stdout + stderr):
        return False
    return True


def contains_errors(text):
    pattern = r"([2-9]|\d{2,})\s+errors?"
    error_flag = bool(re.search(pattern, text))
    return


class DbtError(Exception):
    def __init__(self, message):
        self.message = 'DBT build failed with errors.'

    def __str__(self):
        return self.message


if __name__ == "__main__":
    sbuild()
